@@ -11,6 +11,8 @@
 #include <gemmi/atox.hpp>       // no_sign_atoi
 #include <gemmi/sprintf.hpp>
 #include <gemmi/enumstr.hpp>    // for entity_type_to_string, ...
+#include <gemmi/seqtools.hpp>   // for pdbx_one_letter_code, ...
+#include <gemmi/to_pdb.hpp>     // for use_hetatm
 
 namespace gemmi {
 
@@ -57,7 +59,8 @@ inline std::string qchain(const std::string& s) {
 }
 
 
-void add_cif_atoms(const Structure& st, cif::Block& block, bool use_group_pdb) {
+void add_cif_atoms(const Structure& st, cif::Block& block,
+                   bool use_group_pdb, bool auth_all) {
   // atom list
   cif::Loop& atom_loop = block.init_mmcif_loop("_atom_site.", {
       "id",
@@ -75,9 +78,13 @@ void add_cif_atoms(const Structure& st, cif::Block& block, bool use_group_pdb) {
       "occupancy",
       "B_iso_or_equiv",
       "pdbx_formal_charge",
+      "auth_atom_id",  // optional (tags[15] is removed if !auth_all)
+      "auth_comp_id",  // optional (tags[16] is removed if !auth_all)
       "auth_seq_id",
       "auth_asym_id",
       "pdbx_PDB_model_num"});
+  if (!auth_all)
+    atom_loop.tags.erase(atom_loop.tags.begin() + 15, atom_loop.tags.begin() + 17);
   if (use_group_pdb)
     atom_loop.tags.emplace(atom_loop.tags.begin(), "_atom_site.group_PDB");
   bool has_calc_flag = false;
@@ -88,7 +95,8 @@ void add_cif_atoms(const Structure& st, cif::Block& block, bool use_group_pdb) {
       for (const Residue& res : chain.residues)
         for (const Atom& atom : res.atoms) {
           ++atom_site_count;
-          if (atom.calc_flag != CalcFlag::NotSet)
+          if (atom.calc_flag != CalcFlag::NotSet &&
+              atom.calc_flag != CalcFlag::NoHydrogen)
             has_calc_flag = true;
           if (atom.tls_group_id >= 0)
             has_tls_group_id = true;
@@ -102,11 +110,12 @@ void add_cif_atoms(const Structure& st, cif::Block& block, bool use_group_pdb) {
 
   std::vector<std::string>& vv = atom_loop.values;
   vv.reserve(atom_site_count * atom_loop.tags.size());
-  std::vector<std::pair<int, const Atom*>> aniso;
+  std::vector<std::tuple<int, int, const Atom*>> aniso;
   int serial = 0;
   for (const Model& model : st.models) {
     for (const Chain& chain : model.chains) {
       for (const Residue& res : chain.residues) {
+        bool as_het = use_hetatm(res);
         std::string label_seq_id = res.label_seq.str('.');
         std::string auth_seq_id = res.seqid.num.str();
         std::string entity_id;
@@ -116,7 +125,7 @@ void add_cif_atoms(const Structure& st, cif::Block& block, bool use_group_pdb) {
           entity_id = string_or_dot(res.entity_id);
         for (const Atom& atom : res.atoms) {
           if (use_group_pdb)
-            vv.emplace_back(res.het_flag != 'H' ? "ATOM" : "HETATM");
+            vv.emplace_back(as_het ? "HETATM" : "ATOM");
           vv.emplace_back(std::to_string(++serial));
           vv.emplace_back(atom.element.uname());
           vv.emplace_back(cif::quote(atom.name));
@@ -132,17 +141,22 @@ void add_cif_atoms(const Structure& st, cif::Block& block, bool use_group_pdb) {
           vv.emplace_back(to_str(atom.occ));
           vv.emplace_back(to_str(atom.b_iso));
           vv.emplace_back(atom.charge == 0 ? "?" : std::to_string(atom.charge));
+          if (auth_all) {
+            size_t atom_name_idx = vv.size() - 13;
+            vv.emplace_back(vv[atom_name_idx]);  // auth_atom_id = label_atom_id
+            vv.emplace_back(vv[atom_name_idx + 2]);  // auth_comp_id = label_comp_id
+          }
           vv.emplace_back(auth_seq_id);
           vv.emplace_back(qchain(chain.name));
-          vv.emplace_back(string_or_qmark(model.name));
+          vv.emplace_back(std::to_string(model.num));
           if (has_calc_flag)
-            vv.emplace_back(&".\0d\0c\0dum"[2 * (int) atom.calc_flag]);
+            vv.emplace_back(&".\0.\0d\0c\0dum"[2 * (int) atom.calc_flag]);
           if (has_tls_group_id)
             vv.emplace_back(int_or_qmark(atom.tls_group_id));
           if (st.has_d_fraction)
             vv.emplace_back(to_str(atom.fraction));
           if (atom.aniso.nonzero())
-            aniso.emplace_back(serial, &atom);
+            aniso.emplace_back(serial, model.num, &atom);
         }
       }
     }
@@ -153,17 +167,22 @@ void add_cif_atoms(const Structure& st, cif::Block& block, bool use_group_pdb) {
     cif::Loop& aniso_loop = block.init_mmcif_loop("_atom_site_anisotrop.", {
                                   "id", "type_symbol", "U[1][1]", "U[2][2]",
                                   "U[3][3]", "U[1][2]", "U[1][3]", "U[2][3]"});
+    if (st.models.size() > 1)
+      aniso_loop.tags.push_back("_atom_site_anisotrop.pdbx_PDB_model_num");
     std::vector<std::string>& aniso_val = aniso_loop.values;
     aniso_val.reserve(aniso_loop.tags.size() * aniso.size());
     for (const auto& a : aniso) {
-      aniso_val.emplace_back(std::to_string(a.first));
-      aniso_val.emplace_back(a.second->element.uname());
-      aniso_val.emplace_back(to_str(a.second->aniso.u11));
-      aniso_val.emplace_back(to_str(a.second->aniso.u22));
-      aniso_val.emplace_back(to_str(a.second->aniso.u33));
-      aniso_val.emplace_back(to_str(a.second->aniso.u12));
-      aniso_val.emplace_back(to_str(a.second->aniso.u13));
-      aniso_val.emplace_back(to_str(a.second->aniso.u23));
+      aniso_val.emplace_back(std::to_string(std::get<0>(a)));
+      const Atom* atom = std::get<2>(a);
+      aniso_val.emplace_back(atom->element.uname());
+      aniso_val.emplace_back(to_str(atom->aniso.u11));
+      aniso_val.emplace_back(to_str(atom->aniso.u22));
+      aniso_val.emplace_back(to_str(atom->aniso.u33));
+      aniso_val.emplace_back(to_str(atom->aniso.u12));
+      aniso_val.emplace_back(to_str(atom->aniso.u13));
+      aniso_val.emplace_back(to_str(atom->aniso.u23));
+      if (st.models.size() > 1)
+        aniso_loop.values.push_back(std::to_string(std::get<1>(a)));
     }
   }
 }
@@ -283,6 +302,27 @@ void write_cell_parameters(const UnitCell& cell, cif::ItemSpan& span) {
   span.set_pair("_cell.angle_gamma", to_str(cell.gamma));
 }
 
+bool is_valid_block_name(const std::string& name) {
+  return !name.empty() &&
+         std::all_of(name.begin(), name.end(), [](char c){ return c >= '!' && c <= '~'; });
+}
+
+int get_number_obs(const BasicRefinementInfo& ref) {
+  int nobs = ref.reflection_count;
+  if (nobs == -1 && ref.rfree_set_count >= 0 && ref.work_set_count >= 0)
+    nobs = ref.work_set_count + ref.rfree_set_count;
+  return nobs;
+}
+
+int get_number_work(const BasicRefinementInfo& ref) {
+  int nwork = ref.work_set_count;
+  if (nwork == -1 && ref.rfree_set_count >= 0 && ref.reflection_count >= 0)
+    nwork = ref.reflection_count - ref.rfree_set_count;
+  return nwork;
+}
+
+} // anonymous namespace
+
 void write_ncs_oper(const Structure& st, cif::Block& block) {
   // _struct_ncs_oper (MTRIX)
   if (st.ncs.empty())
@@ -309,13 +349,6 @@ void write_ncs_oper(const Structure& st, cif::Block& block) {
     add_op(op);
 }
 
-bool is_valid_block_name(const std::string& name) {
-  return !name.empty() &&
-         std::all_of(name.begin(), name.end(), [](char c){ return c >= '!' && c <= '~'; });
-}
-
-} // anonymous namespace
-
 void write_struct_conn(const Structure& st, cif::Block& block) {
   // example:
   // disulf1 disulf A CYS 3  SG ? 3 ? 1_555 A CYS 18 SG ? 18 ?  1_555 ? 2.045
@@ -382,6 +415,48 @@ void write_struct_conn(const Structure& st, cif::Block& block) {
       type_loop.add_row({connection_type_to_string((Connection::Type)i)});
 }
 
+void write_cispeps(const Structure& st, cif::Block& block) {
+  cif::Loop* prot_cis_loop = nullptr;
+  int pdbx_id = 0;
+  for (const CisPep& cispep : st.cispeps) {
+    const Model* model = &st.models[0];
+    if (st.models.size() > 1) {
+      model = st.find_model(cispep.model_num);
+      if (!model)
+        continue;
+    }
+    const_CRA cra1 = model->find_cra(cispep.partner_c, true);
+    const_CRA cra2 = model->find_cra(cispep.partner_n, true);
+    if (!cra1.residue || !cra2.residue)
+      continue;
+    if (!prot_cis_loop)
+      prot_cis_loop = &block.init_mmcif_loop("_struct_mon_prot_cis.",
+          {"pdbx_id", "pdbx_PDB_model_num",
+           "label_asym_id", "label_seq_id", "label_comp_id",
+           "auth_asym_id", "auth_seq_id", "pdbx_PDB_ins_code",
+           "pdbx_label_asym_id_2", "pdbx_label_seq_id_2", "pdbx_label_comp_id_2",
+           "pdbx_auth_asym_id_2", "pdbx_auth_seq_id_2", "pdbx_PDB_ins_code_2",
+           "label_alt_id", "pdbx_omega_angle"});
+    auto& v = prot_cis_loop->values;
+    v.emplace_back(std::to_string(++pdbx_id));            // pdbx_id
+    v.emplace_back(std::to_string(model->num));           // pdbx_PDB_model_num
+    v.emplace_back(subchain_or_dot(*cra1.residue));       // label_asym_id
+    v.emplace_back(cra1.residue->label_seq.str('.'));     // label_seq_id
+    v.emplace_back(cra1.residue->name);                   // label_comp_id
+    v.emplace_back(qchain(cispep.partner_c.chain_name));  // auth_asym_id
+    v.emplace_back(cispep.partner_c.res_id.seqid.num.str()); // auth_seq_id
+    v.emplace_back(pdbx_icode(cispep.partner_c.res_id));  // pdbx_PDB_ins_code
+    v.emplace_back(subchain_or_dot(*cra2.residue));       // pdbx_label_asym_id_2
+    v.emplace_back(cra2.residue->label_seq.str('.'));     // pdbx_label_seq_id_2
+    v.emplace_back(cra2.residue->name);                   // pdbx_label_comp_id_2
+    v.emplace_back(qchain(cispep.partner_n.chain_name));  // pdbx_auth_asym_id_2
+    v.emplace_back(cispep.partner_n.res_id.seqid.num.str()); // pdbx_auth_seq_id_2
+    v.emplace_back(pdbx_icode(cispep.partner_n.res_id));  // pdbx_PDB_ins_code_2
+    v.emplace_back(1, cispep.only_altloc ? cispep.only_altloc : '.');
+    v.emplace_back(number_or_qmark(cispep.reported_angle));
+  }
+}
+
 void update_mmcif_block(const Structure& st, cif::Block& block, MmcifOutputGroups groups) {
   if (st.models.empty())
     return;
@@ -437,18 +512,37 @@ void update_mmcif_block(const Structure& st, cif::Block& block, MmcifOutputGroup
                            entity_type_to_string(ent.entity_type)});
   }
 
+  std::map<std::string, std::string> subs_to_strands;
+  if (groups.entity_poly || groups.struct_ref)
+    subs_to_strands = st.models[0].subchain_to_chain();
+
   if (groups.entity_poly) {
-    // In this category we write mandatory tags only.
-    // But if such a file is deposited to the wwPDB it causes:
-    //  ERROR: Missing '_entity_poly.pdbx_strand_id' item.
-    // If pdbx_strand_id was present, OneDep would complain about missing
-    // _entity_poly.pdbx_seq_one_letter_code. So just in case, by default
-    // MmcifOutputGroups::entity_poly is false.
-    cif::Loop& ent_poly_loop = block.init_mmcif_loop("_entity_poly.", {"entity_id", "type"});
+    // If the _entity_poly category is included when depositing to the PDB,
+    // it must contain entity_id, type, pdbx_seq_one_letter_code
+    // and pdbx_strand_id. The last one is not documented as required,
+    // but OneDep shows error when it's not included.
+    cif::Loop& ent_poly_loop = block.init_mmcif_loop("_entity_poly.",
+        {"entity_id", "type", "pdbx_strand_id", "pdbx_seq_one_letter_code"});
     for (const Entity& ent : st.entities)
-      if (ent.entity_type == EntityType::Polymer)
+      if (ent.entity_type == EntityType::Polymer) {
+        if (ent.polymer_type == PolymerType::Unknown)
+          continue;  // not sure what to do here
+        ResidueKind kind = sequence_kind(ent.polymer_type);
+        std::string seq1 = pdbx_one_letter_code(ent.full_sequence, kind);
+        std::string strand_ids;
+        for (const std::string& sub : ent.subchains) {
+          auto strand_id = subs_to_strands.find(sub);
+          if (strand_id != subs_to_strands.end()) {
+            if (!strand_ids.empty())
+              strand_ids += ',';
+            strand_ids += strand_id->second;
+          }
+        }
         ent_poly_loop.add_row({qchain(ent.name),
-                               polymer_type_to_string(ent.polymer_type)});
+                               polymer_type_to_string(ent.polymer_type),
+                               string_or_qmark(strand_ids),
+                               string_or_qmark(seq1)});
+      }
   }
 
   if (groups.struct_ref) { // _struct_ref, _struct_ref_seq
@@ -464,7 +558,6 @@ void update_mmcif_block(const Structure& st, cif::Block& block, MmcifOutputGroup
         "pdbx_auth_seq_align_end", "pdbx_seq_align_end_ins_code"});
     int counter = 0;
     int counter2 = 0;
-    std::map<std::string, std::string> subs = st.models[0].subchain_to_chain();
     for (const Entity& ent : st.entities)
       for (const Entity::DbRef& dbref : ent.dbrefs) {
         ref_loop.add_row({std::to_string(++counter),
@@ -474,8 +567,8 @@ void update_mmcif_block(const Structure& st, cif::Block& block, MmcifOutputGroup
                           string_or_qmark(dbref.accession_code),
                           string_or_qmark(dbref.isoform)});
         for (const std::string& subchain : ent.subchains) {
-          auto strand_id = subs.find(subchain);
-          if (strand_id == subs.end())
+          auto strand_id = subs_to_strands.find(subchain);
+          if (strand_id == subs_to_strands.end())
             continue;
           // DbRef::label_seq_begin/end (_struct_ref_seq.seq_align_beg/end) is
           // not filled in when reading PDB file, so we check it here.
@@ -513,8 +606,18 @@ void update_mmcif_block(const Structure& st, cif::Block& block, MmcifOutputGroup
       for (const std::string& item : ent.full_sequence)
         resnames.insert(Entity::first_mon(item));
     cif::Loop& chem_comp_loop = block.init_mmcif_loop("_chem_comp.", {"id", "type"});
-    for (const std::string& name : resnames)
-      chem_comp_loop.add_row({cif::quote(name), "."});
+    if (!st.shortened_ccd_codes.empty())
+      chem_comp_loop.tags.push_back("_chem_comp.three_letter_code");
+    for (const std::string& name : resnames) {
+      chem_comp_loop.values.push_back(cif::quote(name));
+      chem_comp_loop.values.push_back(".");
+      if (!st.shortened_ccd_codes.empty()) {
+        chem_comp_loop.values.push_back(cif::quote(name));
+        for (const auto& old_new : st.shortened_ccd_codes)
+          if (old_new.second == name)
+            chem_comp_loop.values.back() = old_new.first;
+      }
+    }
   }
 
   if (groups.exptl) {
@@ -637,30 +740,34 @@ void update_mmcif_block(const Structure& st, cif::Block& block, MmcifOutputGroup
                     number_or_qmark(exper.reflections.mean_I_over_sigma),
                     /*number_or_qmark(exper.b_wilson)*/});
     // _reflns_shell
-    cif::Loop& shell_loop = block.init_mmcif_loop("_reflns_shell.", {
-        "pdbx_ordinal",
-        "pdbx_diffrn_id",
-        "d_res_high",
-        "d_res_low",
-        "percent_possible_all",
-        "pdbx_redundancy",
-        "Rmerge_I_obs",
-        "pdbx_Rsym_value",
-        "meanI_over_sigI_obs"});
+    cif::Loop* shell_loop = nullptr;
     n = 0;
     for (const ExperimentInfo& exper : st.meta.experiments) {
       std::string diffrn_id =
         string_or_dot(join_str(exper.diffraction_ids, ","));
-      for (const ReflectionsInfo& shell : exper.shells)
-        shell_loop.add_row({std::to_string(++n),
-                            diffrn_id,
-                            number_or_qmark(shell.resolution_high),
-                            number_or_qmark(shell.resolution_low),
-                            number_or_qmark(shell.completeness),
-                            number_or_qmark(shell.redundancy),
-                            number_or_qmark(shell.r_merge),
-                            number_or_qmark(shell.r_sym),
-                            number_or_qmark(shell.mean_I_over_sigma)});
+      for (const ReflectionsInfo& shell : exper.shells) {
+        if (!shell_loop)
+          shell_loop = &block.init_mmcif_loop("_reflns_shell.", {
+              "pdbx_ordinal",
+              "pdbx_diffrn_id",
+              "d_res_high",
+              "d_res_low",
+              "percent_possible_all",
+              "pdbx_redundancy",
+              "Rmerge_I_obs",
+              "pdbx_Rsym_value",
+              "meanI_over_sigI_obs"});
+
+        shell_loop->add_row({std::to_string(++n),
+                             diffrn_id,
+                             number_or_qmark(shell.resolution_high),
+                             number_or_qmark(shell.resolution_low),
+                             number_or_qmark(shell.completeness),
+                             number_or_qmark(shell.redundancy),
+                             number_or_qmark(shell.r_merge),
+                             number_or_qmark(shell.r_sym),
+                             number_or_qmark(shell.mean_I_over_sigma)});
+      }
     }
   }
 
@@ -672,7 +779,8 @@ void update_mmcif_block(const Structure& st, cif::Block& block, MmcifOutputGroup
         "ls_d_res_high",
         "ls_d_res_low",
         "ls_percent_reflns_obs",
-        "ls_number_reflns_obs"});
+        "ls_number_reflns_obs",
+        "ls_number_reflns_R_work"});
     cif::Loop& analyze_loop = block.init_mmcif_loop("_refine_analyze.", {
         "entry_id",
         "pdbx_refine_id",
@@ -680,24 +788,53 @@ void update_mmcif_block(const Structure& st, cif::Block& block, MmcifOutputGroup
     cif::Loop& restr_loop = block.init_mmcif_loop("_refine_ls_restr.", {
         "pdbx_refine_id", "type",
         "number", "weight", "pdbx_restraint_function", "dev_ideal"});
-    cif::Loop& shell_loop = block.init_mmcif_loop("_refine_ls_shell.", {
+    // _refine_ls_shell
+    std::vector<std::string> shell_tags = {
         "pdbx_refine_id",
         "d_res_high",
         "d_res_low",
         "percent_reflns_obs",
         "number_reflns_obs",
+        "number_reflns_R_work",
         "number_reflns_R_free",
         "R_factor_obs",
         "R_factor_R_work",
-        "R_factor_R_free"});
+        "R_factor_R_free"};
+    bool has_shell_fsc = false;
+    bool has_shell_ffcc = false;
+    bool has_shell_iicc = false;
+    for (const RefinementInfo& ref : st.meta.refinement)
+      for (const BasicRefinementInfo& bin : ref.bins) {
+        if (!std::isnan(bin.fsc_work) || !std::isnan(bin.fsc_free))
+          has_shell_fsc = true;
+        if (!std::isnan(bin.cc_fo_fc_work) || !std::isnan(bin.cc_fo_fc_free))
+          has_shell_ffcc = true;
+        if (!std::isnan(bin.cc_intensity_work) || !std::isnan(bin.cc_intensity_free))
+          has_shell_iicc = true;
+      }
+    if (has_shell_fsc) {
+      shell_tags.push_back("pdbx_fsc_work");
+      shell_tags.push_back("pdbx_fsc_free");
+    }
+    if (has_shell_ffcc) {
+      shell_tags.push_back("correlation_coeff_Fo_to_Fc");
+      shell_tags.push_back("correlation_coeff_Fo_to_Fc_free");
+    }
+    if (has_shell_iicc) {
+      shell_tags.push_back("correlation_coeff_I_to_Fcsqd_work");
+      shell_tags.push_back("correlation_coeff_I_to_Fcsqd_free");
+    }
+    cif::Loop& shell_loop = block.init_mmcif_loop("_refine_ls_shell.", shell_tags);
+
     for (size_t i = 0; i != st.meta.refinement.size(); ++i) {
       const RefinementInfo& ref = st.meta.refinement[i];
-      loop.values.push_back(id);
-      loop.values.push_back(cif::quote(ref.id));
-      loop.values.push_back(number_or_dot(ref.resolution_high));
-      loop.values.push_back(number_or_dot(ref.resolution_low));
-      loop.values.push_back(number_or_dot(ref.completeness));
-      loop.values.push_back(int_or_dot(ref.reflection_count));
+      loop.add_values({id,
+                       cif::quote(ref.id),
+                       number_or_dot(ref.resolution_high),
+                       number_or_dot(ref.resolution_low),
+                       number_or_dot(ref.completeness),
+                       int_or_dot(get_number_obs(ref)),
+                       int_or_qmark(get_number_work(ref))});
       auto add = [&](const std::string& tag, const std::string& val) {
         if (i == 0)
           loop.tags.push_back("_refine." + tag);
@@ -721,11 +858,9 @@ void update_mmcif_block(const Structure& st, cif::Block& block, MmcifOutputGroup
         add("B_iso_mean", number_or_qmark(ref.mean_b));
       if (st.meta.has(&RefinementInfo::aniso_b)) {
         if (i == 0)
-          for (const char* index : {"[1][1]", "[2][2]", "[3][3]",
-                                    "[1][2]", "[1][3]", "[2][3]"})
+          for (const char* index : {"[1][1]", "[2][2]", "[3][3]", "[1][2]", "[1][3]", "[2][3]"})
             loop.tags.push_back(std::string("_refine.aniso_B") + index);
-        const Mat33& t = ref.aniso_b;
-        for (double d : {t[0][0], t[1][1], t[2][2], t[0][1], t[0][2], t[1][2]})
+        for (double d : ref.aniso_b.elements_pdb())
           loop.values.push_back(number_or_qmark(d));
       }
       if (st.meta.has(&RefinementInfo::dpi_blow_r))
@@ -740,11 +875,19 @@ void update_mmcif_block(const Structure& st, cif::Block& block, MmcifOutputGroup
       if (st.meta.has(&RefinementInfo::dpi_cruickshank_rfree))
         add("pdbx_overall_SU_R_free_Cruickshank_DPI",
             number_or_qmark(ref.dpi_cruickshank_rfree));
-      if (st.meta.has(&RefinementInfo::cc_fo_fc))
-        add("correlation_coeff_Fo_to_Fc", number_or_qmark(ref.cc_fo_fc));
+      if (st.meta.has(&RefinementInfo::cc_fo_fc_work))
+        add("correlation_coeff_Fo_to_Fc", number_or_qmark(ref.cc_fo_fc_work));
       if (st.meta.has(&RefinementInfo::cc_fo_fc_free))
         add("correlation_coeff_Fo_to_Fc_free",
             number_or_qmark(ref.cc_fo_fc_free));
+      if (st.meta.has(&RefinementInfo::fsc_work))
+        add("pdbx_average_fsc_work", number_or_qmark(ref.fsc_work));
+      if (st.meta.has(&RefinementInfo::fsc_free))
+        add("pdbx_average_fsc_free", number_or_qmark(ref.fsc_free));
+      if (st.meta.has(&RefinementInfo::cc_intensity_work))
+        add("correlation_coeff_I_to_Fcsqd_work", number_or_qmark(ref.cc_intensity_work));
+      if (st.meta.has(&RefinementInfo::cc_intensity_free))
+        add("correlation_coeff_I_to_Fcsqd_free", number_or_qmark(ref.cc_intensity_free));
       if (!st.meta.solved_by.empty())
         add("pdbx_method_to_determine_struct", string_or_qmark(st.meta.solved_by));
       if (!st.meta.starting_model.empty())
@@ -760,17 +903,29 @@ void update_mmcif_block(const Structure& st, cif::Block& block, MmcifOutputGroup
                             number_or_qmark(restr.weight),
                             string_or_qmark(restr.function),
                             number_or_qmark(restr.dev_ideal)});
-      for (const BasicRefinementInfo& bin : ref.bins)
-        shell_loop.add_row({cif::quote(ref.id),
-                            number_or_dot(bin.resolution_high),
-                            number_or_qmark(bin.resolution_low),
-                            number_or_qmark(bin.completeness),
-                            int_or_qmark(bin.reflection_count),
-                            int_or_qmark(bin.rfree_set_count),
-                            number_or_qmark(bin.r_all),
-                            number_or_qmark(bin.r_work),
-                            number_or_qmark(bin.r_free)});
+      for (const BasicRefinementInfo& bin : ref.bins) {
+        shell_loop.add_values({cif::quote(ref.id),
+                               number_or_dot(bin.resolution_high),
+                               number_or_qmark(bin.resolution_low),
+                               number_or_qmark(bin.completeness),
+                               int_or_qmark(get_number_obs(bin)),
+                               int_or_qmark(get_number_work(bin)),
+                               int_or_qmark(bin.rfree_set_count),
+                               number_or_qmark(bin.r_all),
+                               number_or_qmark(bin.r_work),
+                               number_or_qmark(bin.r_free)});
+        if (has_shell_fsc)
+          shell_loop.add_values({number_or_qmark(bin.fsc_work),
+                                 number_or_qmark(bin.fsc_free)});
+        if (has_shell_ffcc)
+          shell_loop.add_values({number_or_qmark(bin.cc_fo_fc_work),
+                                 number_or_qmark(bin.cc_fo_fc_free)});
+        if (has_shell_iicc)
+          shell_loop.add_values({number_or_qmark(bin.cc_intensity_work),
+                                 number_or_qmark(bin.cc_intensity_free)});
+      }
     }
+    assert(shell_loop.values.size() % shell_loop.tags.size() == 0);
     assert(loop.values.size() % loop.tags.size() == 0);
   }
 
@@ -808,21 +963,20 @@ void update_mmcif_block(const Structure& st, cif::Block& block, MmcifOutputGroup
       }
   }
 
-  if (groups.origx) { // _database_PDB_matrix (ORIGX)
-    if (st.has_origx && !st.origx.is_identity()) {
-      cif::ItemSpan span(block.items, "_database_PDB_matrix.");
-      span.set_pair("_database_PDB_matrix.entry_id", id);
-      std::string tag_mat = "_database_PDB_matrix.origx[0][0]";
-      std::string tag_vec = "_database_PDB_matrix.origx_vector[0]";
-      for (int i = 0; i < 3; ++i) {
-        tag_mat[27] += 1;  // origx[0] -> origx[1] -> origx[2]
-        tag_vec[34] += 1;
-        for (int j = 0; j < 3; ++j) {
-          tag_mat[30] = '1' + j;
-          span.set_pair(tag_mat, to_str(st.origx.mat[i][j]));
-        }
-        span.set_pair(tag_vec, to_str(st.origx.vec.at(i)));
+  bool nontrivial_origx = st.has_origx && !st.origx.is_identity();
+  if (groups.origx && nontrivial_origx) { // _database_PDB_matrix (ORIGX)
+    cif::ItemSpan span(block.items, "_database_PDB_matrix.");
+    span.set_pair("_database_PDB_matrix.entry_id", id);
+    std::string tag_mat = "_database_PDB_matrix.origx[0][0]";
+    std::string tag_vec = "_database_PDB_matrix.origx_vector[0]";
+    for (int i = 0; i < 3; ++i) {
+      tag_mat[27] += 1;  // origx[0] -> origx[1] -> origx[2]
+      tag_vec[34] += 1;
+      for (int j = 0; j < 3; ++j) {
+        tag_mat[30] = '1' + j;
+        span.set_pair(tag_mat, to_str(st.origx.mat[i][j]));
       }
+      span.set_pair(tag_vec, to_str(st.origx.vec.at(i)));
     }
   }
 
@@ -965,24 +1119,37 @@ void update_mmcif_block(const Structure& st, cif::Block& block, MmcifOutputGroup
   if (groups.conn)
     write_struct_conn(st, block);
 
-  if (groups.cis) {  // _struct_mon_prot_cis
-    cif::Loop& prot_cis_loop = block.init_mmcif_loop("_struct_mon_prot_cis.",
-        {"pdbx_id", "pdbx_PDB_model_num", "label_asym_id", "label_seq_id",
-         "auth_asym_id", "auth_seq_id", "pdbx_PDB_ins_code",
-         "label_comp_id", "label_alt_id"});
-    for (const Model& model : st.models)
-      for (const Chain& chain : model.chains)
-        for (const Residue& res : chain.residues)
-          if (res.is_cis)
-            prot_cis_loop.add_row({std::to_string(prot_cis_loop.length()+1),
-                                   model.name, subchain_or_dot(res),
-                                   res.label_seq.str('.'), qchain(chain.name),
-                                   res.seqid.num.str(), pdbx_icode(res),
-                                   res.name, "."});
+  if (groups.cis)  // _struct_mon_prot_cis
+    write_cispeps(st, block);
+
+  // _pdbx_struct_mod_residue (MODRES)
+  if (groups.modres && !st.mod_residues.empty()) {
+    bool use_ccp4_mod_id = false;
+    for (const ModRes& modres : st.mod_residues)
+      if (!modres.mod_id.empty())
+        use_ccp4_mod_id = true;
+    cif::Loop& loop = block.init_mmcif_loop("_pdbx_struct_mod_residue.",
+        {"id", "auth_asym_id", "auth_seq_id", "PDB_ins_code", "auth_comp_id",
+         "label_comp_id", "parent_comp_id", "details"});
+    if (use_ccp4_mod_id)
+      loop.tags.push_back("_pdbx_struct_mod_residue.ccp4_mod_id");
+    int counter = 0;
+    for (const ModRes& modres : st.mod_residues) {
+      loop.add_values({std::to_string(++counter),
+                       qchain(modres.chain_name),
+                       modres.res_id.seqid.num.str(),
+                       pdbx_icode(modres.res_id),
+                       string_or_dot(modres.res_id.name),
+                       string_or_qmark(modres.res_id.name),
+                       string_or_qmark(modres.parent_comp_id),
+                       string_or_qmark(modres.details)});
+      if (use_ccp4_mod_id)
+        loop.values.push_back(string_or_qmark(modres.mod_id));
+    }
   }
 
   // _atom_sites (SCALE)
-  if (groups.scale && (st.has_origx || st.cell.explicit_matrices)) {
+  if (groups.scale && (nontrivial_origx || st.cell.explicit_matrices)) {
     cif::ItemSpan span(block.items, "_atom_sites.");
     span.set_pair("_atom_sites.entry_id", id);
     std::string prefix = "_atom_sites.fract_transf_";
@@ -994,7 +1161,7 @@ void update_mmcif_block(const Structure& st, cif::Block& block, MmcifOutputGroup
       span.set_pair(matrix_idx + "[1]", to_str(frac.mat[i][0]));
       span.set_pair(matrix_idx + "[2]", to_str(frac.mat[i][1]));
       span.set_pair(matrix_idx + "[3]", to_str(frac.mat[i][2]));
-      span.set_pair(prefix + "vector" + idx, to_str(frac.vec.at(i)));
+      span.set_pair(cat(prefix, "vector", idx), to_str(frac.vec.at(i)));
     }
   }
 
@@ -1013,79 +1180,103 @@ void update_mmcif_block(const Structure& st, cif::Block& block, MmcifOutputGroup
   }
 
   if (groups.entity_poly_seq) {
-    // SEQRES from PDB doesn't record microheterogeneity, so if the resulting
-    // cif has unknown("?") _entity_poly_seq.num, it cannot be trusted.
     cif::Loop& poly_loop = block.init_mmcif_loop("_entity_poly_seq.",
-                                           {"entity_id", "num", "mon_id"});
+                                     {"entity_id", "num", "mon_id", "hetero"});
     for (const Entity& ent : st.entities)
-      if (ent.entity_type == EntityType::Polymer)
+      if (ent.entity_type == EntityType::Polymer) {
+        // SEQRES from PDB doesn't record microheterogeneity.
+        std::string hetero_no = ent.reflects_microhetero ? "n" : "?";
         for (size_t i = 0; i != ent.full_sequence.size(); ++i) {
           const std::string& mon_ids = ent.full_sequence[i];
           std::string num = std::to_string(i+1);
           size_t start = 0, end;
           while ((end = mon_ids.find(',', start)) != std::string::npos) {
             poly_loop.add_row({qchain(ent.name), num,
-                               mon_ids.substr(start, end-start)});
+                               mon_ids.substr(start, end-start), "y"});
             start = end + 1;
           }
-          poly_loop.add_row({qchain(ent.name), num, mon_ids.substr(start)});
+          poly_loop.add_row({qchain(ent.name), num, mon_ids.substr(start),
+                             start == 0 ? hetero_no : "y"});
         }
+      }
   }
 
   if (groups.atoms)
-    add_cif_atoms(st, block, groups.group_pdb);
+    add_cif_atoms(st, block, groups.group_pdb, groups.auth_all);
 
-  if (groups.tls && st.meta.has_tls()) {
+  if (groups.tls && st.meta.get_tls_groups() != nullptr) {
+    // pdbx_refine_id doesn't make sense here, but it's required
+    // by the mmCIF spec. In joint refinement, TLS constraints can't be
+    // specific to a dataset, because they constrain the shared model.
     cif::Loop& loop = block.init_mmcif_loop("_pdbx_refine_tls.", {
-        "pdbx_refine_id", "id",
+        "id", "pdbx_refine_id",
+        "origin_x", "origin_y", "origin_z",
         "T[1][1]", "T[2][2]", "T[3][3]", "T[1][2]", "T[1][3]", "T[2][3]",
         "L[1][1]", "L[2][2]", "L[3][3]", "L[1][2]", "L[1][3]", "L[2][3]",
         "S[1][1]", "S[1][2]", "S[1][3]",
         "S[2][1]", "S[2][2]", "S[2][3]",
-        "S[3][1]", "S[3][2]", "S[3][3]",
-        "origin_x", "origin_y", "origin_z"});
+        "S[3][1]", "S[3][2]", "S[3][3]"});
     for (const RefinementInfo& ref : st.meta.refinement)
       for (const TlsGroup& tls : ref.tls_groups) {
-        const Mat33& T = tls.T;
-        const Mat33& L = tls.L;
+        const SMat33<double>& T = tls.T;
+        const SMat33<double>& L = tls.L;
         const Mat33& S = tls.S;
         auto q = number_or_qmark;
-        loop.add_row({cif::quote(ref.id), string_or_dot(tls.id),
-                      q(T[0][0]), q(T[1][1]), q(T[2][2]),
-                      q(T[0][1]), q(T[0][2]), q(T[1][2]),
-                      q(L[0][0]), q(L[1][1]), q(L[2][2]),
-                      q(L[0][1]), q(L[0][2]), q(L[1][2]),
+        loop.add_row({string_or_dot(tls.id), cif::quote(ref.id),
+                      q(tls.origin.x), q(tls.origin.y), q(tls.origin.z),
+                      q(T.u11), q(T.u22), q(T.u33), q(T.u12), q(T.u13), q(T.u23),
+                      q(L.u11), q(L.u22), q(L.u33), q(L.u12), q(L.u13), q(L.u23),
                       q(S[0][0]), q(S[0][1]), q(S[0][2]),
                       q(S[1][0]), q(S[1][1]), q(S[1][2]),
-                      q(S[2][0]), q(S[2][1]), q(S[2][2]),
-                      q(tls.origin.x), q(tls.origin.y), q(tls.origin.z)});
+                      q(S[2][0]), q(S[2][1]), q(S[2][2])});
       }
     cif::Loop& group_loop = block.init_mmcif_loop("_pdbx_refine_tls_group.", {
-        "id", "refine_tls_id", "beg_auth_asym_id", "beg_auth_seq_id",
-        "end_auth_asym_id", "end_auth_seq_id", "selection_details"});
+        "id", "refine_tls_id", "pdbx_refine_id",
+        "beg_auth_asym_id", "beg_auth_seq_id", "beg_PDB_ins_code",
+        "end_auth_asym_id", "end_auth_seq_id", "end_PDB_ins_code",
+        "selection_details"});
     int counter = 1;
     for (const RefinementInfo& ref : st.meta.refinement)
       for (const TlsGroup& tls : ref.tls_groups)
         for (const TlsGroup::Selection& sel : tls.selections)
           group_loop.add_row({std::to_string(counter++),
                               string_or_dot(tls.id),
+                              cif::quote(ref.id),
                               string_or_qmark(sel.chain),
-                              sel.res_begin.num ? sel.res_begin.str() : "?",
+                              sel.res_begin.num.str(),
+                              pdbx_icode(sel.res_begin),
                               string_or_qmark(sel.chain),
-                              sel.res_end.num ? sel.res_end.str() : "?",
+                              sel.res_end.num.str(),
+                              pdbx_icode(sel.res_end),
                               string_or_qmark(sel.details)});
   }
 
   if (groups.software && !st.meta.software.empty()) {
-    cif::Loop& loop = block.init_mmcif_loop("_software.",
-                 {"pdbx_ordinal", "classification", "name", "version", "date"});
+    bool write_all_fields = false;
     for (const SoftwareItem& item : st.meta.software)
-      loop.add_row({
-          std::to_string(item.pdbx_ordinal),
+      if (!item.date.empty() || !item.description.empty() ||
+          !item.contact_author.empty() || !item.contact_author_email.empty())
+        write_all_fields = true;
+    cif::Loop& loop = block.init_mmcif_loop("_software.",
+        {"pdbx_ordinal", "classification", "name", "version"});
+    if (write_all_fields)
+      loop.tags.insert(loop.tags.end(),
+                       {"_software.date", "_software.description",
+                        "_software.contact_author", "_software.contact_author_email"});
+    int ordinal = 0;
+    for (const SoftwareItem& item : st.meta.software) {
+      loop.add_values({
+          std::to_string(++ordinal),
           cif::quote(software_classification_to_string(item.classification)),
           cif::quote(item.name),
-          string_or_dot(item.version),
-          string_or_qmark(item.date)});
+          string_or_dot(item.version)});
+      if (write_all_fields)
+        loop.add_values({
+            string_or_qmark(item.date),
+            string_or_qmark(item.description),
+            string_or_qmark(item.contact_author),
+            string_or_qmark(item.contact_author_email)});
+    }
   }
 }
 
@@ -1113,7 +1304,7 @@ void add_minimal_mmcif_data(const Structure& st, cif::Block& block) {
   write_cell_parameters(st.cell, cell_span);
   block.set_pair("_symmetry.space_group_name_H-M", cif::quote(st.spacegroup_hm));
   write_ncs_oper(st, block);
-  add_cif_atoms(st, block, /*use_group_pdb=*/false);
+  add_cif_atoms(st, block, /*use_group_pdb=*/false, /*auth_all=*/false);
 }
 
 } // namespace gemmi
