@@ -11,21 +11,19 @@
 
 #include <gemmi/fail.hpp>       // for fail
 #include <gemmi/sprintf.hpp>
-#include <gemmi/calculate.hpp>  // for calculate_omega
-#include <gemmi/polyheur.hpp>   // for are_connected
-#include <gemmi/resinfo.hpp>
+#include <gemmi/resinfo.hpp>    // for find_tabulated_residue
 #include <gemmi/util.hpp>
 
 namespace gemmi {
 
 #define WRITE(...) do { \
-    gstb_snprintf(buf, 82, __VA_ARGS__); \
+    snprintf_z(buf, 82, __VA_ARGS__); \
     buf[80] = '\n'; \
     os.write(buf, 81); \
   } while(0)
 
 #define WRITEU(...) do { \
-    gstb_snprintf(buf, 82, __VA_ARGS__); \
+    snprintf_z(buf, 82, __VA_ARGS__); \
     buf[80] = '\n'; \
     for (int i_ = 0; i_ != 80; i_++) \
       if (buf[i_] >= 'a' && buf[i_] <= 'z') buf[i_] -= 0x20; \
@@ -33,15 +31,13 @@ namespace gemmi {
   } while(0)
 
 #define WRITELN(...) do { \
-    int length__ = gstb_snprintf(buf, 82, __VA_ARGS__); \
+    int length__ = snprintf_z(buf, 82, __VA_ARGS__); \
     if (length__ < 80) \
       std::memset(buf + length__, ' ', 80 - length__); \
     buf[80] = '\n'; \
     os.write(buf, 81); \
   } while(0)
 
-namespace {
-
 bool use_hetatm(const Residue& res) {
   if (res.het_flag == 'H')
     return true;
@@ -54,8 +50,10 @@ bool use_hetatm(const Residue& res) {
   return !find_tabulated_residue(res.name).is_standard();
 }
 
+namespace {
+
 // works for non-negative values only
-inline void base36_encode(char* buffer, int width, int value) {
+void base36_encode(char* buffer, int width, int value) {
   const char base36[] = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
   buffer[width] = '\0';
   do {
@@ -67,35 +65,31 @@ inline void base36_encode(char* buffer, int width, int value) {
 }
 
 // based on http://cci.lbl.gov/hybrid_36/
-inline std::array<char,8> encode_serial_in_hybrid36(int serial) {
+std::array<char,8> encode_serial_in_hybrid36(int serial) {
   std::array<char,8> str;
   assert(serial >= 0);
   if (serial < 100000)
-    gstb_sprintf(str.data(), "%5d", serial);
+    to_chars_z(str.data(), str.data() + 8, serial);
   else
-    base36_encode(str.data(), 5, serial - 100000 + 10 * 36 * 36 * 36 * 36);
+    base36_encode(str.data(), 5, serial + (10 * 36 * 36 * 36 * 36 - 100000));
   return str;
 }
 
-// based on http://cci.lbl.gov/hybrid_36/
-inline void encode_seq_num_in_hybrid36(char* str, SeqId::OptionalNum seq_num) {
-  if (*seq_num > -1000 && *seq_num < 10000)
-    gstb_sprintf(str, "%4d", *seq_num);
-  else if (seq_num.has_value())
-    base36_encode(str, 4, *seq_num - 10000 + 10 * 36 * 36 * 36);
-  else
-    std::memcpy(str, "    ", 4);
-}
-
-inline std::array<char,8> write_seq_id(const SeqId& seqid) {
+std::array<char,8> write_seq_id(const SeqId& seqid) {
   std::array<char,8> str;
-  encode_seq_num_in_hybrid36(str.data(), *seqid.num);
-  str[4] = seqid.icode;
-  str[5] = '\0';
+  char* ptr = str.data();
+  if (*seqid.num > -1000 && *seqid.num < 10000) {
+    ptr = to_chars_z(ptr, ptr + 5, *seqid.num);
+  } else if (seqid.num.has_value()) {
+    base36_encode(ptr, 4, *seqid.num + (10 * 36 * 36 * 36 - 10000));
+    ptr += 4;
+  }
+  *ptr++ = seqid.icode;
+  *ptr = '\0';
   return str;
 }
 
-inline const char* find_last_break(const char *str, int max_len) {
+const char* find_last_break(const char *str, int max_len) {
   int last_break = 0;
   for (int i = 0; i < max_len; i++) {
     if (str[i] == '\0')
@@ -108,8 +102,8 @@ inline const char* find_last_break(const char *str, int max_len) {
 
 // Write record with possible continuation lines, with the format:
 // 1-6 record name, 8-10 continuation, 11-lastcol string.
-inline void write_multiline(std::ostream& os, const char* record_name,
-                            const std::string& text, int lastcol) {
+void write_multiline(std::ostream& os, const char* record_name,
+                     const std::string& text, int lastcol) {
   if (text.empty())
     return;
   char buf[88]; // a few bytes extra, just in case
@@ -124,16 +118,7 @@ inline void write_multiline(std::ostream& os, const char* record_name,
   }
 }
 
-inline void write_cryst1(const Structure& st, std::ostream& os) {
-  char buf[88];
-  const UnitCell& cell = st.cell;
-  WRITE("CRYST1%9.3f%9.3f%9.3f%7.2f%7.2f%7.2f %-11s%4s          ",
-        cell.a, cell.b, cell.c, cell.alpha, cell.beta, cell.gamma,
-        st.spacegroup_hm.empty() ? "P 1" : st.spacegroup_hm.c_str(),
-        st.get_info("_cell.Z_PDB").c_str());
-}
-
-inline void write_ncs_op(const NcsOp& op, std::ostream& os) {
+void write_ncs_op(const NcsOp& op, std::ostream& os) {
   char buf[88];
   for (int i = 0; i < 3; ++i) {
     WRITE("MTRIX%d %3.3s%10.6f%10.6f%10.6f %14.5f    %-21c", i+1,
@@ -142,18 +127,7 @@ inline void write_ncs_op(const NcsOp& op, std::ostream& os) {
   }
 }
 
-inline void write_ncs(const Structure& st, std::ostream& os) {
-  if (st.ncs.empty())
-    return;
-  auto identity = st.info.find("_struct_ncs_oper.id");
-  if (identity != st.info.end() &&
-      !in_vector_f([&](const NcsOp& op) { return op.id == identity->second; }, st.ncs))
-    write_ncs_op(NcsOp{identity->second, true, {}}, os);
-  for (const NcsOp& op : st.ncs)
-    write_ncs_op(op, os);
-}
-
-inline void write_remarks(const Structure& st, std::ostream& os) {
+void write_remarks(const Structure& st, std::ostream& os) {
   char buf[88];
   if (st.resolution > 0) {
     WRITE("%-80s", "REMARK   2");
@@ -234,8 +208,8 @@ inline void write_remarks(const Structure& st, std::ostream& os) {
   }
 }
 
-inline void write_chain_atoms(const Chain& chain, std::ostream& os,
-                              int& serial, PdbWriteOptions opt) {
+void write_chain_atoms(const Chain& chain, std::ostream& os,
+                       int& serial, PdbWriteOptions opt) {
   char buf[88];
   buf[0] = '\0';
   if (chain.name.length() > 2)
@@ -243,6 +217,7 @@ inline void write_chain_atoms(const Chain& chain, std::ostream& os,
   for (const Residue& res : chain.residues) {
     bool as_het = use_hetatm(res);
     for (const Atom& a : res.atoms) {
+      serial = opt.preserve_serial ? a.serial : serial + 1;
       //  1- 6  6s  record name
       //  7-11  5d  integer serial
       // 12     1   -
@@ -263,11 +238,11 @@ inline void write_chain_atoms(const Chain& chain, std::ostream& os,
       // 73-76      segment identifier, left-justified (non-standard)
       // 77-78  2s  element symbol, right-justified
       // 79-80  2s  charge
-      WRITE("%-6s%5s %-4.4s%c%3s"
-            "%2s%5s   %8.3f%8.3f%8.3f"
-            "%6.2f%6.2f      %-4.4s%2s%c%c",
+      int written_bytes = snprintf_z(buf, 82,
+            "%-6s%5s %-4.4s%c%3.3s"
+            "%2s%5s   %8.3f%8.3f%8.3f",
             as_het ? "HETATM" : "ATOM",
-            encode_serial_in_hybrid36(++serial).data(),
+            encode_serial_in_hybrid36(serial).data(),
             a.padded_name().c_str(),
             a.altloc ? std::toupper(a.altloc) : ' ',
             res.name.c_str(),
@@ -277,7 +252,19 @@ inline void write_chain_atoms(const Chain& chain, std::ostream& os,
             // if they originally had one digit more and that digit was 5.
             a.pos.x > -5e-4 && a.pos.x < 0 ? 0 : a.pos.x + 1e-10,
             a.pos.y > -5e-4 && a.pos.y < 0 ? 0 : a.pos.y + 1e-10,
-            a.pos.z > -5e-4 && a.pos.z < 0 ? 0 : a.pos.z + 1e-10,
+            a.pos.z > -5e-4 && a.pos.z < 0 ? 0 : a.pos.z + 1e-10);
+      if GEMMI_UNLIKELY(written_bytes > 54) {
+        // The only items expected to overflow above are the coordinates,
+        // if the integer part of the number exceeds 5 characters.
+        // This happens when something goes wrong and the model is far from
+        // the origin. Such a model should be shifted; it can't be written it
+        // in a spec-conforming format: Real(8.3). Here we overwrite the last
+        // digits - trimming is better than overflowing the line.
+        snprintf_z(buf+38, 82-38, "%8.3f", a.pos.y);
+        snprintf_z(buf+46, 82-46, "%8.3f", a.pos.z);
+      }
+      snprintf_z(buf+54, 82-54,
+            "%6.2f%6.2f      %-4.4s%2s%c%c",
             // Occupancy is stored as single prec, but we know it's <= 1,
             // so no precision is lost even if it had 6 digits after dot.
             a.occ + 1e-6,
@@ -291,14 +278,16 @@ inline void write_chain_atoms(const Chain& chain, std::ostream& os,
             // Sometimes PDB files have explicit 0s (5M05); we ignore them.
             a.charge ? a.charge > 0 ? '0'+a.charge : '0'-a.charge : ' ',
             a.charge ? a.charge > 0 ? '+' : '-' : ' ');
+      buf[80] = '\n';
+      os.write(buf, 81);
       if (a.aniso.nonzero()) {
         // re-using part of the buffer
         std::memcpy(buf, "ANISOU", 6);
         const double eps = 1e-6;
-        gstb_snprintf(buf+28, 43, "%7.0f%7.0f%7.0f%7.0f%7.0f%7.0f",
-                      a.aniso.u11*1e4 + eps, a.aniso.u22*1e4 + eps,
-                      a.aniso.u33*1e4 + eps, a.aniso.u12*1e4 + eps,
-                      a.aniso.u13*1e4 + eps, a.aniso.u23*1e4 + eps);
+        snprintf_z(buf+28, 43, "%7.0f%7.0f%7.0f%7.0f%7.0f%7.0f",
+                   a.aniso.u11*1e4 + eps, a.aniso.u22*1e4 + eps,
+                   a.aniso.u33*1e4 + eps, a.aniso.u12*1e4 + eps,
+                   a.aniso.u13*1e4 + eps, a.aniso.u23*1e4 + eps);
         buf[28+42] = ' ';
         buf[80] = '\n';
         os.write(buf, 81);
@@ -312,8 +301,8 @@ inline void write_chain_atoms(const Chain& chain, std::ostream& os,
       if (opt.numbered_ter) {
         // re-using part of the buffer in the middle, e.g.:
         // TER    4153      LYS B 286
-        gstb_snprintf(buf, 82, "TER   %5s",
-                      encode_serial_in_hybrid36(++serial).data());
+        snprintf_z(buf, 82, "TER   %5s",
+                   encode_serial_in_hybrid36(++serial).data());
         std::memset(buf+11, ' ', 6);
         std::memset(buf+28, ' ', 52);
         buf[80] = '\n';
@@ -325,33 +314,21 @@ inline void write_chain_atoms(const Chain& chain, std::ostream& os,
   }
 }
 
-inline void write_atoms(const Structure& st, std::ostream& os,
-                        PdbWriteOptions opt) {
-  char buf[88];
-  for (const Model& model : st.models) {
-    int serial = 0;
-    if (st.models.size() > 1) {
-      // according to the spec model name in mmCIF may not be numeric
-      std::string name = model.name;
-      for (char c : name)
-        if (!std::isdigit(c)) {
-          name = std::to_string(&model - &st.models[0] + 1);
-          break;
-        }
-      WRITE("MODEL %8s %65s", name.c_str(), "");
-    }
-    for (const Chain& chain : model.chains)
-      write_chain_atoms(chain, os, serial, opt);
-    if (st.models.size() > 1)
-      WRITE("%-80s", "ENDMDL");
-  }
-}
+} // anonymous namespace
+
+void write_pdb(const Structure& st, std::ostream& os, PdbWriteOptions opt) {
+  // check if structure can be written as pdb
+  for (const gemmi::Model& model : st.models)
+    for (const gemmi::Chain& chain : model.chains)
+      if (chain.name.size() > 2)
+        gemmi::fail("chain name too long for the PDB format: " + chain.name);
 
-inline void write_header(const Structure& st, std::ostream& os,
-                         PdbWriteOptions opt) {
   const std::string& entry_id = st.get_info("_entry.id");
+  const char* entry_id_4 = entry_id.size() <= 4 ? entry_id.c_str() : "";
   char buf[88];
-  { // header line
+
+  // HEADER
+  if (!opt.minimal_file) {
     const char* months = "JANFEBMARAPRMAYJUNJULAUGSEPOCTNOVDEC???";
     const std::string& date =
       st.get_info("_pdbx_database_status.recvd_initial_deposition_date");
@@ -367,28 +344,32 @@ inline void write_header(const Structure& st, std::ostream& os,
       WRITEU("HEADER    %-40.40s%-9s   %-18.18s",
              keywords.c_str(), pdb_date.c_str(), entry_id.c_str());
   }
-  write_multiline(os, "TITLE", st.get_info("_struct.title"), 80);
-  write_multiline(os, "KEYWDS", st.get_info("_struct_keywords.text"), 79);
-  std::string expdta = st.get_info("_exptl.method");
-  if (expdta.empty())
-    expdta = join_str(st.meta.experiments, "; ",
-                      [](const ExperimentInfo& e) { return e.method; });
-  write_multiline(os, "EXPDTA", expdta, 79);
-  if (st.models.size() > 1)
-    WRITE("NUMMDL    %-6zu %63s", st.models.size(), "");
 
-  if (!st.raw_remarks.empty()) {
-    for (const std::string& line : st.raw_remarks) {
-      os << line;
-      if (line.empty() || line.back() != '\n')
-        os << '\n';
+  // TITLE, KEYWDS, EXPDTA, NUMMDL, REMARK
+  if (!opt.minimal_file) {
+    write_multiline(os, "TITLE", st.get_info("_struct.title"), 80);
+    write_multiline(os, "KEYWDS", st.get_info("_struct_keywords.text"), 79);
+    std::string expdta = st.get_info("_exptl.method");
+    if (expdta.empty())
+      expdta = join_str(st.meta.experiments, "; ",
+                        [](const ExperimentInfo& e) { return e.method; });
+    write_multiline(os, "EXPDTA", expdta, 79);
+    if (st.models.size() > 1)
+      WRITE("NUMMDL    %-6zu %63s", st.models.size(), "");
+
+    if (!st.raw_remarks.empty()) {
+      for (const std::string& line : st.raw_remarks) {
+        os << line;
+        if (line.empty() || line.back() != '\n')
+          os << '\n';
+      }
+    } else {
+      write_remarks(st, os);
     }
-  } else {
-    write_remarks(st, os);
   }
 
   // DBREF[12], SEQRES
-  if (!st.models.empty() && opt.seqres_records) {
+  if (opt.seqres_records && !st.models.empty()) {
     std::vector<const Entity*> entity_list;
     entity_list.reserve(st.models[0].chains.size());
     for (const Chain& ch : st.models[0].chains) {
@@ -410,7 +391,6 @@ inline void write_header(const Structure& st, std::ostream& os,
     for (size_t i = 0; i != entity_list.size(); ++i)
       if (const Entity* entity = entity_list[i]) {
         const Chain& ch = st.models[0].chains[i];
-        const char* dbref_entry_id = entry_id.size() <= 4 ? entry_id.c_str() : "";
         for (const Entity::DbRef& dbref : entity->dbrefs) {
           bool short_record = *dbref.db_end.num < 100000 &&
                               dbref.accession_code.size() < 9 &&
@@ -422,11 +402,11 @@ inline void write_header(const Structure& st, std::ostream& os,
               begin = polymer.label_seq_id_to_auth(dbref.label_seq_begin);
               end = polymer.label_seq_id_to_auth(dbref.label_seq_end);
             }
-          gstb_snprintf(buf, 82, "DBREF  %4s%2s %5s %5s %-6s  ",
-                        dbref_entry_id, ch.name.c_str(),
-                        write_seq_id(begin).data(),
-                        write_seq_id(end).data(),
-                        dbref.db_name.c_str());
+          snprintf_z(buf, 82, "DBREF  %4s%2s %5s %5s %-6s  ",
+                     entry_id_4, ch.name.c_str(),
+                     write_seq_id(begin).data(),
+                     write_seq_id(end).data(),
+                     dbref.db_name.c_str());
           if (dbref.db_name == "PDB" && dbref.id_code == entry_id) {
             // PDB uses self-reference for fragments that don't have real
             // reference. No idea why. In such case the same begin/end is used.
@@ -435,19 +415,19 @@ inline void write_header(const Structure& st, std::ostream& os,
             end = dbref.db_end;
           }
           if (short_record) {
-            gstb_snprintf(buf+33, 82-33, "%-8s %-12s %5d%c %5d%c            \n",
-                          dbref.accession_code.c_str(), dbref.id_code.c_str(),
-                          *begin.num, begin.icode, *end.num, end.icode);
+            snprintf_z(buf+33, 82-33, "%-8s %-12s %5d%c %5d%c            \n",
+                       dbref.accession_code.c_str(), dbref.id_code.c_str(),
+                       *begin.num, begin.icode, *end.num, end.icode);
           } else {
             buf[5] = '1';  // -> DBREF1
-            gstb_snprintf(buf+33, 82-33, "              %-33s\n",
-                          dbref.id_code.c_str());
+            snprintf_z(buf+33, 82-33, "              %-33s\n",
+                       dbref.id_code.c_str());
           }
           buf[80] = '\n';
           os.write(buf, 81);
           if (!short_record)
             WRITE("DBREF2 %4s%2s     %-22s     %10d  %10d             ",
-                  dbref_entry_id, ch.name.c_str(),
+                  entry_id_4, ch.name.c_str(),
                   dbref.accession_code.c_str(), *begin.num, *end.num);
         }
       }
@@ -459,12 +439,10 @@ inline void write_header(const Structure& st, std::ostream& os,
         int col = 0;
         for (const std::string& monomers : entity->full_sequence) {
           if (col == 0)
-            gstb_snprintf(buf, 82, "SEQRES%4d%2s%5zu %62s\n",
-                          ++row, ch.name.c_str(),
-                          entity->full_sequence.size(), "");
-          size_t end = monomers.find(',');
-          if (end == std::string::npos)
-            end = monomers.length();
+            snprintf_z(buf, 82, "SEQRES%4d%2s%5zu %62s\n",
+                       ++row, ch.name.c_str(),
+                       entity->full_sequence.size(), "");
+          size_t end = std::min(monomers.find(','), std::min(monomers.length(), (size_t)3));
           std::memcpy(buf + 18 + 4*col + 4-end, monomers.c_str(), end);
           if (++col == 13) {
             os.write(buf, 81);
@@ -478,14 +456,53 @@ inline void write_header(const Structure& st, std::ostream& os,
       }
   }
 
-  if (!st.helices.empty()) {
+  if (!opt.minimal_file) {
+    // MODRES
+    for (const ModRes& modres : st.mod_residues) {
+      WRITEU("MODRES %4s %3.3s%2s %5s %3s  %-41.41s  %-8.8s\n",
+            entry_id_4,
+            modres.res_id.name.c_str(),
+            modres.chain_name.c_str(),
+            write_seq_id(modres.res_id.seqid).data(),
+            modres.parent_comp_id.c_str(),
+            modres.details.c_str(),
+            modres.mod_id.c_str());
+    }
+
+    // HET
+    if (!st.models.empty()) {
+      auto full_ccd_code = [&](const std::string& name) -> std::string {
+        for (const auto& mapping : st.shortened_ccd_codes)
+          if (mapping.second == name)
+            return "real CCD code: " + mapping.first;
+        return "";
+      };
+      for (const Chain& chain : st.models[0].chains)
+        for (const Residue& res : chain.residues)
+          if (use_hetatm(res) && !find_tabulated_residue(res.name).is_water()) {
+            WRITE("HET    %3s %2s%5s %6zu     %-51s\n",
+                  res.name.c_str(),
+                  chain.name.c_str(),
+                  write_seq_id(res.seqid).data(),
+                  res.atoms.size(),
+                  res.name[0] == '~' ? full_ccd_code(res.name).c_str() : "");
+          }
+    }
+  }
+
+  // HETNAM - but it's used only for tilde-hetnam extension
+  for (const auto& mapping : st.shortened_ccd_codes)
+    WRITE("HETNAM     %3s %55s %-9s\n", mapping.second.c_str(), "", mapping.first.c_str());
+
+  // HELIX
+  if (!opt.minimal_file && !st.helices.empty()) {
     int counter = 0;
     for (const Helix& helix : st.helices) {
       if (++counter == 10000)
         counter = 0;
       // According to the PDB spec serial number can be from 1 to 999.
       // Here, we allow for up to 9999 helices by using columns 7 and 11.
-      gstb_snprintf(buf, 82, "HELIX %4d%4d %3s%2s %5s %3s%2s %5s%2d %35d    \n",
+      snprintf_z(buf, 82, "HELIX %4d%4d %3.3s%2s %5s %3.3s%2s %5s%2d %35d    \n",
             counter, counter,
             helix.start.res_id.name.c_str(), helix.start.chain_name.c_str(),
             write_seq_id(helix.start.res_id.seqid).data(),
@@ -499,15 +516,16 @@ inline void write_header(const Structure& st, std::ostream& os,
     }
   }
 
-  if (!st.sheets.empty()) {
+  // SHEET
+  if (!opt.minimal_file && !st.sheets.empty()) {
     for (const Sheet& sheet : st.sheets) {
       int strand_counter = 0;
       for (const Sheet::Strand& strand : sheet.strands) {
         const AtomAddress& a2 = strand.hbond_atom2;
         const AtomAddress& a1 = strand.hbond_atom1;
         // H-bond atom names are expected to be O and N
-        WRITE("SHEET%5d %3.3s%2zu %3s%2s%5s %3s%2s%5s%2d  %-3s%3s%2s%5s "
-              " %-3s%3s%2s%5s          ",
+        WRITE("SHEET%5d %3.3s%2zu %3.3s%2s%5s %3.3s%2s%5s%2d  %-3s%3.3s%2s%5s "
+              " %-3s%3.3s%2s%5s          ",
               ++strand_counter, sheet.name.c_str(), sheet.strands.size(),
               strand.start.res_id.name.c_str(), strand.start.chain_name.c_str(),
               write_seq_id(strand.start.res_id.seqid).data(),
@@ -537,7 +555,7 @@ inline void write_header(const Structure& st, std::ostream& os,
                                                        cra2.atom->pos, con.asu);
           if (++counter == 10000)
             counter = 0;
-          WRITE("SSBOND%4d %3s%2s %5s %5s%2s %5s %28s %6s %5.2f  ",
+          WRITE("SSBOND%4d %3.3s%2s %5s   %3.3s%2s %5s %28s %6s %5.2f  ",
              counter,
              cra1.residue->name.c_str(), cra1.chain->name.c_str(),
              write_seq_id(cra1.residue->seqid).data(),
@@ -548,7 +566,7 @@ inline void write_header(const Structure& st, std::ostream& os,
     }
 
     // LINK  (note: uses only the first model and primary conformation)
-    if (!st.models.empty() && opt.link_records) {
+    if (opt.link_records) {
       for (const Connection& con : st.connections)
         if (con.type == Connection::Covale || con.type == Connection::MetalC ||
             con.type == Connection::Unknown) {
@@ -571,8 +589,8 @@ inline void write_header(const Structure& st, std::ostream& os,
           // to be applied to the atom." But all files from wwPDB have
           // 1555 not blank, so here we also write 1555,
           // except for LINKR (Refmac variant of LINK).
-          gstb_snprintf(buf, 82, "LINK        %-4s%c%3s%2s%5s   "
-                "            %-4s%c%3s%2s%5s  %6s %6s %5s  \n",
+          snprintf_z(buf, 82, "LINK        %-4s%c%3.3s%2s%5s   "
+                "            %-4s%c%3.3s%2s%5s  %6s %6s %5s  \n",
                 cra1.atom ? cra1.atom->padded_name().c_str() : "",
                 cra1.atom && cra1.atom->altloc ? std::toupper(cra1.atom->altloc) : ' ',
                 cra1.residue->name.c_str(),
@@ -584,94 +602,116 @@ inline void write_header(const Structure& st, std::ostream& os,
                 con.partner2.chain_name.c_str(),
                 write_seq_id(cra2.residue->seqid).data(),
                 "1555", im_pdb_symbol.c_str(), im_dist_str.c_str());
-          if (opt.use_linkr && !con.link_id.empty()) {
-            buf[4] = 'R';  // LINK -> LINKR
-            if (im_same_asu)
-              std::memset(buf+58, ' ', 14); // erase symmetry
+          if (im_same_asu && !con.link_id.empty())
+            std::memset(buf+58, ' ', 14); // erase symmetry
+          if ((opt.use_linkr ||opt.use_link_id) && !con.link_id.empty()) {
+            if (opt.use_linkr)
+              buf[4] = 'R';  // LINK -> LINKR
             // overwrite distance with link_id
-            gstb_snprintf(buf+72, 82-72, "%-8s\n", con.link_id.c_str());
+            if (opt.use_linkr || opt.use_link_id)
+              snprintf_z(buf+72, 82-72, "%-8s\n", con.link_id.c_str());
           }
           buf[80] = '\n';
           os.write(buf, 81);
         }
-
     }
 
-    // CISPEP (note: uses only the first conformation)
-    if (!st.models.empty() && opt.cispep_records) {
+    // CISPEP
+    if (opt.cispep_records) {
       int counter = 0;
-      for (const Model& model : st.models)
-        for (const Chain& chain : model.chains)
-          for (const Residue& res : chain.residues)
-            if (res.is_cis) {
-              const Residue* next = chain.next_residue(res);
-              if (next && are_connected(res, *next, PolymerType::PeptideL)) {
-                if (++counter == 10000)
-                  counter = 0;
-                WRITE("CISPEP%4d %3s%2s %5s   %3s%2s %5s %9s %12.2f %20s",
-                      counter,
-                      res.name.c_str(), chain.name.c_str(),
-                      write_seq_id(res.seqid).data(),
-                      next->name.c_str(), chain.name.c_str(),
-                      write_seq_id(next->seqid).data(),
-                      st.models.size() > 1 ? model.name.c_str() : "0",
-                      deg(calculate_omega(res, *next)),
-                      "");
-              }
-            }
+      for (const CisPep& cispep : st.cispeps) {
+        WRITE("CISPEP%4d %3.3s%2s %5s   %3.3s%2s %5s %9d %12.2f %20s",
+              ++counter,
+              cispep.partner_c.res_id.name.c_str(),
+              cispep.partner_c.chain_name.c_str(),
+              write_seq_id(cispep.partner_c.res_id.seqid).data(),
+              cispep.partner_n.res_id.name.c_str(),
+              cispep.partner_n.chain_name.c_str(),
+              write_seq_id(cispep.partner_n.res_id.seqid).data(),
+              st.models.size() > 1 ? cispep.model_num : 0,
+              std::isnan(cispep.reported_angle) ? 0. : cispep.reported_angle,
+              "");
+        if (counter == 9999)
+          counter = 0;
+      }
     }
   }
 
-  if(opt.cryst1_record)
-    write_cryst1(st, os);
-  if (st.has_origx && !st.origx.is_identity()) {
+  // CRYST1
+  if (opt.cryst1_record) {
+    WRITE("CRYST1%9.3f%9.3f%9.3f%7.2f%7.2f%7.2f %-11s%4s          ",
+          st.cell.a, st.cell.b, st.cell.c, st.cell.alpha, st.cell.beta, st.cell.gamma,
+          st.spacegroup_hm.empty() ? "P 1" : st.spacegroup_hm.c_str(),
+          st.get_info("_cell.Z_PDB").c_str());
+  }
+
+  // ORIGX1/2/3
+  if (!opt.minimal_file && st.has_origx && !st.origx.is_identity()) {
     for (int i = 0; i < 3; ++i)
       WRITE("ORIGX%d %13.6f%10.6f%10.6f %14.5f %24s", i+1,
             st.origx.mat[i][0], st.origx.mat[i][1], st.origx.mat[i][2],
             st.origx.vec.at(i), "");
   }
-  if (st.cell.explicit_matrices) {
+
+  // SCALE1/2/3
+  if (!opt.minimal_file && st.cell.explicit_matrices) {
     for (int i = 0; i < 3; ++i)
       // We add a small number to avoid negative 0.
       WRITE("SCALE%d %13.6f%10.6f%10.6f %14.5f %24s", i+1,
             st.cell.frac.mat[i][0] + 1e-15, st.cell.frac.mat[i][1] + 1e-15,
             st.cell.frac.mat[i][2] + 1e-15, st.cell.frac.vec.at(i) + 1e-15, "");
   }
-  write_ncs(st, os);
+
+  // MTRIX1/2/3
+  if (!st.ncs.empty()) {
+    auto identity = st.info.find("_struct_ncs_oper.id");
+    if (identity != st.info.end() &&
+        !in_vector_f([&](const NcsOp& op) { return op.id == identity->second; }, st.ncs))
+      write_ncs_op(NcsOp{identity->second, true, {}}, os);
+    for (const NcsOp& op : st.ncs)
+      write_ncs_op(op, os);
+  }
+
+  // MODEL, ATOM, HETATM, TER, ENDMDL
+  if (opt.atom_records) {
+    for (const Model& model : st.models) {
+      int serial = 0;
+      if (st.models.size() > 1)
+        WRITE("MODEL %8d %65s", model.num, "");
+      for (const Chain& chain : model.chains)
+        write_chain_atoms(chain, os, serial, opt);
+      if (st.models.size() > 1)
+        WRITE("%-80s", "ENDMDL");
+    }
+  }
+
+  // CONECT
+  if (opt.conect_records) {
+    auto num_str = [](const std::vector<int>& nums, size_t i) {
+      return i < nums.size() ? encode_serial_in_hybrid36(nums[i]) : std::array<char,8>{};
+    };
+    for (const auto& num_pair : st.conect_map)
+      for (size_t i = 0; i < num_pair.second.size(); i += 4)
+        WRITE("CONECT%5s%5s%5s%5s%5s%50s",
+              encode_serial_in_hybrid36(num_pair.first).data(),
+              encode_serial_in_hybrid36(num_pair.second[i]).data(),
+              num_str(num_pair.second, i+1).data(),
+              num_str(num_pair.second, i+2).data(),
+              num_str(num_pair.second, i+3).data(),
+              "");
+  }
+
+  // END
+  if (opt.end_record)
+    WRITE("%-80s", "END");
 }
 
-inline void check_if_structure_can_be_written_as_pdb(const Structure& st) {
-  for (const gemmi::Model& model : st.models)
-    for (const gemmi::Chain& chain : model.chains)
-      if (chain.name.size() > 2)
-        gemmi::fail("chain name too long for the PDB format: " + chain.name);
-}
-
-} // anonymous namespace
-
-std::string make_pdb_headers(const Structure& st) {
-  check_if_structure_can_be_written_as_pdb(st);
+std::string make_pdb_string(const Structure& st, PdbWriteOptions opt) {
   std::ostringstream os;
-  write_header(st, os, PdbWriteOptions());
+  write_pdb(st, os, opt);
   return os.str();
 }
 
-void write_pdb(const Structure& st, std::ostream& os, PdbWriteOptions opt) {
-  check_if_structure_can_be_written_as_pdb(st);
-  write_header(st, os, opt);
-  write_atoms(st, os, opt);
-  char buf[88];
-  WRITE("%-80s", "END");
-}
-
-void write_minimal_pdb(const Structure& st, std::ostream& os,
-                       PdbWriteOptions opt) {
-  check_if_structure_can_be_written_as_pdb(st);
-  write_cryst1(st, os);
-  write_ncs(st, os);
-  write_atoms(st, os, opt);
-}
-
 #undef WRITE
 #undef WRITEU
 
